/******************************************************************************
 * (c) Copyright 2002-2006, 1060 Research Ltd
 *
 * This Software is licensed to You, the licensee, for use under the terms of
 * the 1060 Public License v1.0. Please read and agree to the 1060 Public
 * License v1.0 [www.1060research.com/license] before using or redistributing
 * this software.
 *
 * In summary the 1060 Public license has the following conditions.
 * A. You may use the Software free of charge provided you agree to the terms
 * laid out in the 1060 Public License v1.0
 * B. You are only permitted to use the Software with components or applications
 * that provide you with OSI Certified Open Source Code [www.opensource.org], or
 * for which licensing has been approved by 1060 Research Limited.
 * You may write your own software for execution by this Software provided any
 * distribution of your software with this Software complies with terms set out
 * in section 2 of the 1060 Public License v1.0
 * C. You may redistribute the Software provided you comply with the terms of
 * the 1060 Public License v1.0 and that no warranty is implied or given.
 * D. If you find you are unable to comply with this license you may seek to
 * obtain an alternative license from 1060 Research Limited by contacting
 * license@1060research.com or by visiting www.1060research.com
 *
 * NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC
 * LICENSE V1.0 FOR DETAILS
 *
 * THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ
 * THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license
 *
 * File:          $RCSfile$
 * Version:       $Name$ $Revision$
 * Last Modified: $Date$
 *****************************************************************************/
package org.netkernel.layer0.nkf;

import org.netkernel.urii.IMetaRepresentation;

/**
 * Convenience helper methods for sourcing and sinking resources. Implemented over INKFBasicContext
 * @author  tab
 */
public interface INKFRequestContext extends INKFBasicContext, INKFLocale, INKFParameters
{
	/** Source a resource
	 * @param aIdentifier
	 * @exception NKFException
	 */
	Object source(String aIdentifier) throws NKFException;
	
	/** Source a resource from its URI
	 * @param aIdentifier
	 * @param aRepresentationClass
	 * @exception NKFException
	 */
	<T> T source(String aURI, Class<T> aRepresentationClass) throws NKFException;
	
	<T> INKFResponseReadOnly<T> sourceForResponse(String aIdentifier, Class<T> aRepresentationClass) throws NKFException;
	INKFResponseReadOnly<Object> sourceForResponse(String aIdentifier) throws NKFException;
	
	<T> T sourcePrimary(Class<T> aRepresentationClass) throws NKFException;
	
	<T> INKFResponseReadOnly<T> sourcePrimaryForResponse(Class<T> aRepresentationClass) throws NKFException;
	
	/** Sink a representation to the resource identified by aIdentifier
	 * @param aIdentifier
	 * @param aRepresentation
	 * @exception NKFException
	 */
	void sink(String aIdentifier, Object aRepresentation) throws NKFException;
	
	/** Check for the existence of a resource for the given Identifier
	 */
	boolean exists(String aIdentifier) throws NKFException;
	
	/** Delete a resource for the given Identifier
	 * @return true if it was deleted
	 */
	boolean delete(String aIdentifier) throws NKFException;

	/** Transrept a internally created resource representation into alternative form.
	 * There must be a registered Transreptor capable of performing the
	 * transrepresentation. - Do not use to transrept external resources
	 * source these using one of the methods which take an representation class
	 * parameter.
	 * @param aSource An internally generated resource
	 * @param aRepresentationClass The class of the desired representation for the resource
	 * @return An representation of class aRepresentationClass for resource aSource
	 * @exception NKFException thrown if no transreptor available or transreptor fails
	 */
	<T> T transrept(Object aSource, Class<T> aRepresentationClass) throws NKFException;
	
	<T> INKFResponseReadOnly<T> transreptForResponse(Object aSource, Class<T> aRepresentationClass) throws NKFException;
	
	/** Request the creation of a new resource
	 * @param aIdentifierBase A base identifier that can identify which accessor to target
	 * the request at
	 * @param aOptionalSource Optional resource representation that new resource can
	 * use to seed its state.
	 * @return The identifier of the newly created resource
	 * @exception NKFException Thrown if no accessor could be located or it
	 * didn't support the NEW operation or the failed for some other accessor
	 * specific reason.
	 */
	String requestNew(String aIdentifierBase, Object aOptionalSource) throws NKFException;	
	
	/** Request metadata for endpoint/representation/prototype identified by
	 * specified identifier within current space
	 * @param aIdentifier
	 * @return a subclass of IMetaRepresentation
	 * @throws NKFException if identifier cannot be resolved
	 */
	IMetaRepresentation meta(String aIdentifier) throws NKFException;
	
	/** Request metadata for endpoint which may be qualified by the request
	 * @param aRequest
	 * @return
	 * @throws NKFException if identifier cannot be resolved
	 */
	IMetaRepresentation meta(INKFRequest aRequest) throws NKFException;
	
}