package org.netkernel.layer0.nkf.impl;

import org.netkernel.container.IKernel;
import org.netkernel.layer0.nkf.*;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.IResponse;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.urii.INetKernelException;
import org.netkernel.urii.ISpace;

public class NKFTransportContextImpl extends NKFContextImpl
{
	public NKFTransportContextImpl(IKernel aKernel, ISpace aSpace, Object aTransportInstance)
	{	super(aKernel,aSpace,null,aTransportInstance,null);
	}

	public NKFTransportContextImpl(IKernel aKernel, ISpace aContext, Object aTransportInstance, INKFParametersControl aParameters, IRequestScopeLevel aRootRequestScope)
	{	super(aKernel,aContext,null,aTransportInstance,aParameters);
		setRequestScope(aRootRequestScope);
	}

	public IResponse issueKernelRequest(IRequest aRequest) throws NKFException
	{	try
		{	return mKernel.synchronousExternalRequest(aRequest, getEndpoint(), getSpace());
		} catch (INetKernelException e)
		{	throw new NKFException(e);
		}
	}
	
	public INKFAsyncRequestHandle issueAsyncRequest(INKFRequest aRequest) throws NKFException
	{	if (aRequest==null) throw new IllegalArgumentException(getErrorMessage("EX_NKF_SUBREQ_NULL"));
		IRequest kernelReq = ((NKFRequestImpl)aRequest).getKernelRequest();
		NKFAsyncRequestHandleImpl subRequest = new NKFAsyncRequestHandleImpl((NKFRequestImpl)aRequest,this);
		kernelReq=RequestFactory.cloneWithRequestor(kernelReq,subRequest);
		mKernel.onAsyncExternalRequest(kernelReq, getEndpoint(), getSpace());
		return subRequest;	
	}
	
	public IRequest getThisKernelRequest()
	{	return null;
	}
	
	public void handleAsyncResponse(IResponse aResponse)
	{	//do nothing here as a transport has no one to return responses too
	}
}