package org.netkernel.layer0.util;
import java.util.*;

public class TupleList
{
	private List mList;
	private final int mTupleLength;
	
	/** Construct a new TupleList
	 * @param aTupleLength number of items each tuple has
	 * @param aInitialSize the initial allocation of tuples though this can grow. If zero then the tuple is fixed as empty
	 */
	public TupleList(int aTupleLength, int aInitialSize)
	{	mTupleLength=aTupleLength;
		if (aInitialSize>0)
		{	mList=new ArrayList(mTupleLength*aInitialSize);
		}
		else
		{	mList=Collections.EMPTY_LIST;
		}
	}
	
	/**
	 * @param aValues
	 */
	public void put(Object... aValues)
	{	if (aValues.length==mTupleLength)
		{	for (int i=0; i<mTupleLength; i++)
			{	mList.add(aValues[i]);
			}
		}
		else throw new IllegalArgumentException();
	}
	
	public Object getValue(int aIndex, int aTuple)
	{	int index=aIndex*mTupleLength+aTuple;
		return mList.get(index);
	}
	
	public int size()
	{	return mList.size()/mTupleLength;
	}
	
	public int getTupleLength()
	{	return mTupleLength;
	}
}
