/******************************************************************************
  (c) Copyright 2002-2009, 1060 Research Ltd                                   

  This Software is licensed to You, the licensee, for use under the terms of   
  the 1060 Public License v1.0. Please read and agree to the 1060 Public       
  License v1.0 [www.1060research.com/license] before using or redistributing   
  this software.                                                               

  In summary the 1060 Public license has the following conditions.             
  A. You may use the Software free of charge provided you agree to the terms   
  laid out in the 1060 Public License v1.0                                     
  B. You are only permitted to use the Software with components or applications
  that provide you with OSI Certified Open Source Code [www.opensource.org], or
  for which licensing has been approved by 1060 Research Limited.              
  You may write your own software for execution by this Software provided any  
  distribution of your software with this Software complies with terms set out 
  in section 2 of the 1060 Public License v1.0                                 
  C. You may redistribute the Software provided you comply with the terms of   
  the 1060 Public License v1.0 and that no warranty is implied or given.       
  D. If you find you are unable to comply with this license you may seek to    
  obtain an alternative license from 1060 Research Limited by contacting       
  license@1060research.com or by visiting www.1060research.com                 

  NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC  
  LICENSE V1.0 FOR DETAILS                                                     

  THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ     
  THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license      

  File:          $RCSfile$
  Version:       $Name$ $Revision$
  Last Modified: $Date$
 *****************************************************************************/
package org.netkernel.layer0.util;

import java.io.*;
import java.lang.management.ManagementFactory;
import java.lang.management.MemoryPoolMXBean;
import java.lang.management.MemoryType;
import java.security.MessageDigest;
import java.util.*;

import org.netkernel.container.IMessages;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSFactory;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.IResponse;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.scheduler.TraceDebugCapture;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.INetKernelException;
import org.netkernel.urii.IResolution;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.ISpaceWithIdentity;
import org.netkernel.util.ExclusionSet;
import org.w3c.dom.Document;

/** General utility methods
 * @author tab
 */
public class Utils extends org.netkernel.util.Utils
{
	private static int sUnique;
	private static MemoryPoolMXBean sOldGen;
	
	static
	{	List<MemoryPoolMXBean> pools= ManagementFactory.getMemoryPoolMXBeans();
		for (MemoryPoolMXBean pool : pools)
		{	if (pool.getType()==MemoryType.HEAP && pool.isUsageThresholdSupported())
			{	sOldGen=pool;
			}
		}	
	}
	
	public static MemoryPoolMXBean getOldGen()
	{	return sOldGen;
	}
	
	/** copy an input stream to an outputsteam and close streams when finished
	 * @throws IOException if there are any problems
	 */
	public static void pipe(InputStream aInput, OutputStream aOutput) throws IOException
	{	byte b[] = new byte[256];
		int c;
		try
		{	while ( (c=aInput.read(b))>0 )
			{	aOutput.write(b,0,c);
			}
		}
		finally
		{	try
			{	aInput.close();
			}
			finally
			{	aOutput.close();
			}
		}
	}
	
	/** copy an input stream to an outputsteam without closing output stream
	 * @throws IOException if there are any problems
	 */
	public static void pipeNoClose(InputStream aInput, OutputStream aOutput) throws IOException
	{	byte b[] = new byte[256];
		int c;
		try
		{	while ( (c=aInput.read(b))>0 )
			{	aOutput.write(b,0,c);
			}
		}
		finally
		{	aInput.close();
		}
	}
	
	/** copy an reader to a writer without closing the writer
	 * @throws IOException if there are any problems
	 */
	public static void pipeNoClose(Reader aInput, Writer aOutput) throws IOException
	{	char b[] = new char[256];
		int c;
		try
		{	while ( (c=aInput.read(b))>0 )
			{	aOutput.write(b,0,c);
			}
		}
		finally
		{	aInput.close();
		}
	}
	
	
	
	/** ensure windows style slashes become URI style slashes */
	public static String fixSlash(String aURI)
	{	return aURI.replace('\\', '/');
	}
	
	/** Appends a URI string to the string buffer escaping an reserved characters
	 * ( / - . _ ~ and alphanumeric are unreserved) 
	 * @param sb
	 * @param aURI
	 */
	public static void appendUnreservedURIChar(StringBuilder sb, String aURI)
	{	int length=aURI.length();
		for (int i=0; i<length; i++)
		{	char c=aURI.charAt(i);
			if (c=='/'||c=='-'||c=='.'||c=='_'||c=='~'||Character.isLetterOrDigit(c))
			{	sb.append(c);
			}
			else
			{	sb.append('%');
				sb.append(Integer.toHexString(c));
			}
		}
	}
	
	/** Decodes a identifiers escaped with %
	 * @param aInput
	 * @return
	 */
	public static String decode(String aInput)
	{	String result=aInput;
		try
		{	StringBuilder sb=new StringBuilder(aInput.length());
			byte[] decodeBuffer = new byte[aInput.length()/3];
			int length = aInput.length();
			for (int i=0; i<length; i++)
			{	char c = aInput.charAt(i);
				if (c=='%')
				{	int p=0;
					do
					{	i++;
						int c1 = Utils.toHex(aInput.charAt(i++));
						int c2 = Utils.toHex(aInput.charAt(i++));
						decodeBuffer[p++]=(byte)((c1<<4) + c2);
					} while (i<length && aInput.charAt(i)=='%');
					sb.append(new String(decodeBuffer,0,p,"UTF-8"));
					i--;
				}
				else
				{	sb.append(c);
				}
			}			
			if (sb.length()!=aInput.length())
			{	result=sb.toString();
			}
		}
		catch (IllegalArgumentException e) {;}
		catch (UnsupportedEncodingException e) {;}
		return result;
	}
	
	
	/** Converts a single digit, upper case or number to an integer
	 * @param aChar
	 * @exception IllegalArgumentException thrown if aChar is not a valid hex character
	 * @return
	 */
	public static int toHex(char aChar)
	{	int result;
		if (aChar<='9' && aChar>='0')
		{	result = aChar-'0';
		}
		else if (aChar<='F' && aChar>='A')
		{	result= aChar-('A'-10);
		}
		else if (aChar<='f' && aChar>='a')
		{	result= aChar-('a'-10);
		}
		else
		{	throw new IllegalArgumentException();
		}
		return result;
	}
	
	private static char fromHex(int aDigit)
	{	char result=0;
		if (aDigit<10)
		{	result = (char)(aDigit+'0');
		}
		else 
		{	result= (char)(aDigit-10+'A');
		}
		return result;
	}	
	
	/** Parses a an input stream which contains a line delimited strings. 
	 * Each line that doesn't start with a # is put into the result. 
	 * @param aIS
	 * @return
	 * @throws IOException
	 */
	public static ArrayList ArrayListFromStream(InputStream aIS) throws IOException
	{
		String line;
        ArrayList result = new ArrayList();
        BufferedReader in = new BufferedReader(new InputStreamReader(aIS));
        try
        {  if (!in.ready())
	        {	throw new IOException();
	        }
	
	        while ((line = in.readLine()) != null)
	        {	if (!line.startsWith("#") && line.trim().length()>0 )
	        	{	result.add(line);
	        	}
	        }
        }
        finally
        {	in.close();
        }
        return result;
    }
	
	/** Splits a string with the same semantics as a string tokenizer but places the
	 * tokens into an array
	 * @param aInput the input string
	 * @param aSplitChars the characters to use as splitters
	 * @return an array of tokens
	 */
	public static String[] splitString(String aInput, String aSplitChars)
	{	StringTokenizer st = new StringTokenizer(aInput,aSplitChars);
		String [] result = new String[st.countTokens()];
		for (int i=0; i<result.length; i++)
		{	result[i]=st.nextToken();
		}
		return result;
	}
	
	/*
	public static String generateRandomId(int aLength)
	{	char[] chars = new char[aLength];
		Random r = new Random();
		for (int i=0; i<chars.length; i++)
		{	chars[i] = ((char)r.nextInt(26));
			chars[i]+='A';
		}
		return new String(chars);
	}
	*/
	
	
	/** Delegate the resolve request into an "imported" space. This method issues the sub-request direct to
	 * the delegate
	 * @param aDelegate the delegate we want to attempt to resolve the request
	 * @param aResolveRequest the resolve request we want resolving
	 * @return an IResolution from the delegate or null if no resolution possible
	 * @throws NetKernelException thrown if we fail to successfully issue the request
	 */
	public static IResolution delegateResolutionInto(ISpace aDelegate, IRequest aResolveRequest) throws INetKernelException
	{	return delegateResolutionInto(aDelegate,aResolveRequest,null);
	}
	
	/** Delegate the resolve request into an "imported" space. This method issues the sub-request direct to
	 * the delegate
	 * @param aDelegate the delegate we want to attempt to resolve the request
	 * @param aResolveRequest the resolve request we want resolving
	 * @param aScopeToResolveIn this request context is used as the context for the resolve
	 * @return an IResolution from the delegate or null if no resolution possible
	 * @throws NetKernelException thrown if we fail to successfully issue the request
	 */
	public static IResolution delegateResolutionInto(ISpace aDelegate, IRequest aResolveRequest, IRequestScopeLevel aScopeToResolveIn) throws INetKernelException
	{	IResolution result=null;
		ExclusionSet eliminatedSpaces = (ExclusionSet)aResolveRequest.getValue(IRequest.TYPE_RESOLVE_EXCLUDED_SPACES);
		if (!eliminatedSpaces.contains(aDelegate))
		{	IRequest toResolveIn = (IRequest)aResolveRequest.getValue(IRequest.TYPE_RESOLVE_REQUEST);
			TraceDebugCapture debug = (TraceDebugCapture)aResolveRequest.getValue(IRequest.TYPE_RESOLVE_DEBUG);
			IRequestScopeLevel rc;
			if (aScopeToResolveIn==null)
			{	rc = RequestScopeLevelImpl.createNonDurableScopeLevel(aDelegate,toResolveIn.getRequestScope());
			}
			else
			{	rc=aScopeToResolveIn;
			}
			IRequest toResolveOut=RequestFactory.cloneWithScope(toResolveIn,rc);
			SynchronousRequestor sr = new SynchronousRequestor(aDelegate);
			IRequest req = RequestFactory.createResolveRequest(toResolveOut,eliminatedSpaces,debug,aResolveRequest,sr);
			IResponse resp=sr.synchronousRequest(req);
			Object rep=resp.getRepresentation();
			if (rep instanceof IResolution)
			{	result=(IResolution)rep;
			}
		}
		return result;
	}
	
	public static String shortClassName(String aClassname)
	{	String result=aClassname;
		int i=result.lastIndexOf('.');
		if (i>0)
		{	result=result.substring(i+1);
		}
		i=result.lastIndexOf('$');
		if (i>0)
		{	result=result.substring(i+1);
		}
		return result;
	}
	/** Returns the name of the class without package
	 */
	public static String shortClassName(Class aClass)
	{	return shortClassName(aClass.getName());
	}
	
	/** A toString representation of an array [<item1>, <item2>]
	 * @param aArray array to convert to string
	 * @param aExpectedLength the expected length of output
	 * @return the generated string
	 */
	public static String arrayToString(Object[] aArray, int aExpectedLength)
	{	StringBuffer sb=new StringBuffer(aExpectedLength);
		sb.append('[');
		for (int i=0; i<aArray.length; i++)
		{	if (i>0)
			{	sb.append(", ");
			}
			if (aArray[i]!=null)
			{	sb.append(aArray[i].toString());
			}
			else
			{	sb.append("null");
			}
		}
		sb.append(']');
		return sb.toString();
	}
	
	/** Delegate the received request (from aContext) to the given space. This helper method is useful
	 *  for mappers and proxies
	 * @param aDelegate the space to delegate to
	 * @param aContext the INKFRequestContext containing the incoming request
	 * @throws NKFException thrown if the subrequest fails
	 */
	public static void delegateRequestInto(ISpace aDelegate, INKFRequestContext aContext) throws NKFException
	{	INKFRequest reqOut=aContext.getThisRequest().getIssuableClone();
		reqOut.injectRequestScope(aDelegate);
		INKFResponseReadOnly resp=aContext.issueRequestForResponse(reqOut);
		aContext.createResponseFrom(resp);	
	}
	
	/** Determine if the given request is circular, i.e. will it result in a stack overflow
	 * @param aRequest
	 * @return
	 */
	public static boolean isCircular(IRequest aRequest)
	{	boolean result=false;
		ISpace ourSpace=aRequest.getRequestScope().getSpace();
		IRequest req=aRequest.getParent();
		while (req!=null)
		{	if (ourSpace==req.getRequestScope().getSpace() && aRequest.getVerb()==req.getVerb())
			{	if (aRequest.getIdentifier().equals(req.getIdentifier()) && aRequest.getRepresentationClass()==req.getRepresentationClass())
				{	//additionally check that full request scopes match
					IRequestScopeLevel rsla=aRequest.getRequestScope().getParent();
					IRequestScopeLevel rsli=req.getRequestScope().getParent();
					boolean match=true;
					while (rsla!=null && rsli!=null)
					{	if (rsla.getSpace()!=rsli.getSpace())
						{	match=false;
							break;
						}
						rsla=rsla.getParent();
						rsli=rsli.getParent();
					}
					if (match)
					{	result=true;
						break;
					}
				}
			}
			req=req.getParent();
		}
		return result;
	}
	
	
	
	public static InputStream getInputStreamFrom(IBinaryStreamRepresentation aInput, int aExpectedSize) throws IOException
	{	ByteArrayOutputStream baos = new ByteArrayOutputStream(aExpectedSize);
		aInput.write(baos);
		baos.flush();
		InputStream is = new ByteArrayInputStream(baos.toByteArray());
		return is;
	}
	
	public static int scatterHash(int aHash)
	{	aHash += ~(aHash << 9);
		aHash ^=  (aHash >>> 14);
		aHash +=  (aHash << 4);
		aHash ^=  (aHash >>> 10);
		return aHash;
	}
	
	public static String toDataURI(IBinaryStreamRepresentation aRepresentation, String aMimetype) throws IOException
	{
		if (aMimetype==null)
		{	aMimetype="content/unknown";
		}

		String result;
		if (aRepresentation instanceof IReadableBinaryStreamRepresentation)
		{	//efficient way if it's readable
			IReadableBinaryStreamRepresentation rbs = (IReadableBinaryStreamRepresentation)aRepresentation;
			InputStream is=new Base64.InputStream(rbs.getInputStream(),Base64.ENCODE|Base64.GZIP|Base64.DONT_BREAK_LINES);
			Reader r = new BufferedReader(new InputStreamReader(is,"US-ASCII"));
			StringBuilder sb=new StringBuilder(rbs.getContentLength()+64);
			sb.append("data:");
			sb.append(aMimetype);
			sb.append(";base64,");
			while(r.ready())
			{	sb.append((char)r.read());
			}
			result=sb.toString();
		}
		else
		{	//less efficient way if it's not readable
			ByteArrayOutputStream baos=new ByteArrayOutputStream(1024);
			OutputStream os=new Base64.OutputStream(baos,Base64.ENCODE|Base64.GZIP|Base64.DONT_BREAK_LINES);
			aRepresentation.write(os);
			String data=baos.toString("US-ASCII");
			StringBuilder sb=new StringBuilder(data.length()+64);
			sb.append("data:");
			sb.append(aMimetype);
			sb.append(";base64,");
			sb.append(data);
			result=sb.toString();
		}
		return result;
	}
	
	public static String generateUniqueEndpointIdentifier(ISpace aSpace, IEndpoint aEndpoint, String aId)
	{
		int ctxHash;
		if (aSpace instanceof ISpaceWithIdentity)
		{	ISpaceWithIdentity ctx = (ISpaceWithIdentity)aSpace;
			ctxHash=ctx.getIdentifier().toString().hashCode()^ctx.getVersion().toString().hashCode();
		}
		else
		{	ctxHash=aSpace.toString().hashCode();
		}
		ctxHash=scatterHash(ctxHash);
		
		String epId = aEndpoint.getClass().getName();
		int epHash = scatterHash(epId.hashCode());
		if (aId!=null)
		{	epHash^=aId.hashCode();
		}
		
		StringBuilder sb = new StringBuilder();
		sb.append("ep:");
		appendHash(sb,ctxHash,6);
		sb.append('-');
		appendHash(sb,epHash,6);		
		
		return sb.toString();
	}
	
	private static void appendHash(StringBuilder aSB, int aHash, int aSize)
	{ 	if (aHash<0)
		{	aHash=-aHash;
		}
		final int N=36;
		for (int i=0; i<aSize; i++)
		{	int d = aHash%N;
			int c;
			if (d<26)
			{	c='A'+d;
			}
			else
			{	c='0'+d-26;
			}
			aSB.append((char)c);
			aHash/=N;
		}
	}
	
	/**
	 *  Replaces the parameters in the buf (identified by %int), by the String representation of
	 *  the value.
	 *  @param buf  The buffer containing the message to be edited
	 *  @param aValues  The parmeter values to be inserted into buf.
	 */
	public static void replaceParameters(StringBuilder aMessage, Object... aValues)
	{	int location[] = new int[aValues.length];
		for (int i=aValues.length-1; i>=0; i--)
		{	String param = "%" + (i+1);
			int j = aMessage.indexOf(param);
			location[i]=j;
		}
		for (int i=aValues.length-1; i>=0; i--)
		{	int l=location[i];
			try
			{	String sub;
				if (aValues[i]!=null)
				{	sub=aValues[i].toString();
				}
				else
				{	sub="<null>";
				}
				if (l>=0)
				{	aMessage.replace(l,l+2, sub);
					int diff=sub.length()-2;
					for (int j=i-1; j>=0; j--)
					{	if (location[j]>l)
						{	location[j]+=diff;
						}
					}
				}
			}
			catch (Throwable t)
			{	
			}
		}
	}
	
	public static NKFException createFormattedNKFException(String aId, String aMessage, String aLocation, Throwable aCause, IMessages aMessages, Object... aMessageParams)
	{	String id=(aId!=null)?aMessages.format(aId):null;
		String message=null;
		if (aMessage!=null)
		{	message=aMessages.format(aMessage,aMessageParams);
		}	
		return new NKFException(id,message,aLocation,aCause);		
	}
	
	/** Standard parsing a valency string. default value is 1 for zero length string, * is max int value, other values are parsed as int string. faillure to
	 *  parse results in -1
	 * @param aValency
	 * @return
	 */
	public static int parseValency(String aValency, int aDefault)
	{	int result;
		if (aValency.length()==0)
		{	result=aDefault;
		}
		else
		{	if (aValency.equals("*"))
			{	result = Integer.MAX_VALUE;
			}
			else
			{	try
				{	result = Integer.parseInt(aValency);
				}
				catch (NumberFormatException e)
				{	result = -1;
				}
			}
		}
		return result;
	}
	
	public static int parseVerbList(String aVerbString) throws IllegalArgumentException
	{
		StringTokenizer st = new StringTokenizer(aVerbString,",");
		int verbs=0;
		while (st.hasMoreTokens())
		{	String verbString = st.nextToken();
			int verb = RequestFactory.parseVerb(verbString);
			if (verb==0)
			{	throw new IllegalArgumentException();
			}
			verbs|=verb;
		}
		return verbs;
	}

	public static String globStringToRegexString(String aGlob)
	{	StringBuilder buffer = new StringBuilder(aGlob.length()*2);
		char [] chars = aGlob.toCharArray();
		for (int i = 0; i < chars.length; ++i)
		{	if (chars[i] == '*')
			{	buffer.append(".*");
			}
			else if (chars[i] == '?')
			{	buffer.append(".");
			}
			else if ("+()^$.{}[]|\\".indexOf(chars[i]) != -1)
			{	buffer.append('\\').append(chars[i]);
			}
			else
			{	buffer.append(chars[i]);
			}
		}
		return buffer.toString();
	}
	
	public static String toStringOrNull(Object aObject)
	{	return aObject==null?null:aObject.toString();
	}
	
	public static Document sourceAsDOMAvoidingTransrept(INKFRequestContext aContext, String aIdentifier) throws Exception
	{	return (Document)sourceResponseAsDOMAvoidingTransrept(aContext,aIdentifier)[0];
	}
	
	public static Object[] sourceResponseAsDOMAvoidingTransrept(INKFRequestContext aContext, String aIdentifier) throws Exception
	{	Document result;
		INKFResponseReadOnly resp=aContext.sourceForResponse(aIdentifier);
		Object representation=resp.getRepresentation();
		if(representation instanceof Document)
		{	result=(Document)representation;			
		}
		else if(representation instanceof IReadableBinaryStreamRepresentation)
		{	IReadableBinaryStreamRepresentation rbs=(IReadableBinaryStreamRepresentation)representation;
			result=XMLUtils.parse(rbs.getInputStream());
		}
		else if (representation instanceof IHDSNode)
		{	IHDSNode hds=(IHDSNode)representation;
			result=HDSFactory.toDOM(hds);
		}
		else if (representation instanceof String)
		{	String s=(String)representation;
			InputStream is = new ByteArrayInputStream(s.getBytes("UTF-8"));
			result=XMLUtils.parse(is);
		}
		else
		{	resp=aContext.sourceForResponse(aIdentifier, Document.class);
			result=(Document)resp.getRepresentation();
		}
		return new Object[] {result,resp};
	}	
	
	/** java.io.File.toURI() has a bug, it doesn't handle spaces in filenames correctly, it should convert them to %20
	 * @param aFile
	 * @return
	 */
	public static String fileAsURIString(File aFile)
	{
		String s=aFile.toURI().toString();
		if (s.contains(" "))
		{	s=s.replaceAll("\\ ", "%20");
		}
		return s;
	}
	
	/** Format given period as a human readable string
	 * @param aPeriod
	 * @param doSeconds if true display seconds
	 * @return
	 */
	public static String formatPeriod(long aPeriod, boolean doSeconds)
	{
		long period=aPeriod;
	    long second=1000;
	    long minute=second*60;
	    long hour=minute*60;
	    long day=hour*24;
	    long week=day*7;
	    long year=day*365;
	    String message="";
	    long years=period/year;
	    if (years>0)
	    {   message+=years;
	        message+=(years==1)?" year, ":" years, ";
	        period-=years*year;
	    }
	    long weeks=period/week;
	    if (weeks>0 || years>0)
	    {   message+=weeks;
	        message+=(weeks==1)?" week, ":" weeks, ";
	        period-=weeks*week;
	    }
	    long days=period/day;
	    if (days>0 || weeks>0 || years>0)
	    {   message+=days;
	        message+=(days==1)?" day, ":" days, ";
	        period-=days*day;
	    }
	    long hours=period/hour;
	    if (hours>0 || days>0 || weeks>0 || years>0)
	    {   message+=hours;
	        message+=(hours==1)?" hour, ":" hours, ";
	        period-=hours*hour;
	    }
	    long minutes=period/minute;
	    {   message+=minutes;
	        message+=(minutes==1)?" minute":" minutes";
	        period-=minutes*minute;
	    }
	    if (doSeconds)
	    {   message+=", ";
	        long seconds=period/second;
	        {   message+=seconds;
	            message+=" seconds";
	        }
	    }
	    return message;
	}	
	
	static final byte[] HEX_CHAR_TABLE = {
	    (byte)'0', (byte)'1', (byte)'2', (byte)'3',
	    (byte)'4', (byte)'5', (byte)'6', (byte)'7',
	    (byte)'8', (byte)'9', (byte)'a', (byte)'b',
	    (byte)'c', (byte)'d', (byte)'e', (byte)'f'
	  };    

	  public static String toHexString(byte[] ba) 
	  {		byte[] hex = new byte[2 * ba.length];
			int index = 0;
			for (byte b : ba) {
			  int v = b & 0xFF;
			  hex[index++] = HEX_CHAR_TABLE[v >>> 4];
			  hex[index++] = HEX_CHAR_TABLE[v & 0xF];
			}
			String result=null;
			try
			{	result=new String(hex, "ASCII");
			}
			catch(UnsupportedEncodingException e)
			{	//Can't happen but dump it anyway in case...
				e.printStackTrace();				
			}
			return result;
	  }
	  
	public static byte[] fromHexString(String s)
	{	byte[]b =new byte[s.length()/2];
		int k=0;
		for(int i=0; i<s.length()/2; i++)
		{	Integer j=Integer.parseInt(s.substring(2*i, (2*i)+2), 16);
			b[k]=j.byteValue();
			k++;
		}
		return b;
	}
	
	/** Split a string into a list of substrings each of which attempts to have a maximum length
	 * of aMaxLength. If a substring cannot be split into shorter substrings it is possible
	 * that some may be longer that aMaxLength.
	 * @param aText The text to split
	 * @param aMaxLength The maximum length of each substring
	 * @param aDelimiters A string containing the set of delimiters
	 * @return the list of substrings
	 */
	public static List<String> splitIntoRows(String aText, int aMaxLength, String aDelimiters)
	{
		StringTokenizer st=new StringTokenizer(aText,aDelimiters,true);
		List<String> result=new ArrayList<String>();
		StringBuilder sb=new StringBuilder(aMaxLength);
		while (st.hasMoreTokens())
		{	String part=st.nextToken();
			if (part.length()+sb.length()<=aMaxLength)
			{	sb.append(part);
			}
			else if (sb.length()==0)
			{	sb.append(part);
			}
			else
			{	result.add(sb.toString());
				sb.setLength(0);
				sb.append(part);
			}
		}
		if (sb.length()>0)
		{	result.add(sb.toString());
		}
		return result;
	}	
	
}
