/******************************************************************************
 * (c) Copyright 2002,2003, 1060 Research Ltd
 *
 * This Software is licensed to You, the licensee, for use under the terms of
 * the 1060 Public License v1.0. Please read and agree to the 1060 Public
 * License v1.0 [www.1060research.com/license] before using or redistributing
 * this software.
 *
 * In summary the 1060 Public license has the following conditions.
 * A. You may use the Software free of charge provided you agree to the terms
 * laid out in the 1060 Public License v1.0
 * B. You are only permitted to use the Software with components or applications
 * that provide you with OSI Certified Open Source Code [www.opensource.org], or
 * for which licensing has been approved by 1060 Research Limited.
 * You may write your own software for execution by this Software provided any
 * distribution of your software with this Software complies with terms set out
 * in section 2 of the 1060 Public License v1.0
 * C. You may redistribute the Software provided you comply with the terms of
 * the 1060 Public License v1.0 and that no warranty is implied or given.
 * D. If you find you are unable to comply with this license you may seek to
 * obtain an alternative license from 1060 Research Limited by contacting
 * license@1060research.com or by visiting www.1060research.com
 *
 * NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC
 * LICENSE V1.0 FOR DETAILS
 *
 * THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ
 * THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license
 *
 * File:          $RCSfile$
 * Version:       $Name$ $Revision$
 * Last Modified: $Date$
 *****************************************************************************/
package org.netkernel.layer0.util;

import org.netkernel.layer0.representation.ByteArrayRepresentation;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.w3c.dom.*;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import java.io.*;
import java.util.Properties;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

/**
 * A set of low level DOM manipulation methods
 * @author  tab
 */
public class XMLUtils {
	
	/** return a new empty document
	 */
	public static Document newDocument() throws ParserConfigurationException
	{	
		Thread t=Thread.currentThread();
		ClassLoader cl=t.getContextClassLoader();
		t.setContextClassLoader(XMLUtils.class.getClassLoader());
		try
		{	
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();	
			DocumentBuilder builder = factory.newDocumentBuilder();			
			return builder.newDocument();
		}
		finally
		{	t.setContextClassLoader(cl);
		}		
	}	
	
	/** escape the given string into a form safe to embed into an XML stream
	 */
	public static String escape(String aXML)
	{	return org.netkernel.util.XMLUtils.escape(aXML);
	}
	
	
	public static String getPathFor(Node aNode) {
		StringBuilder result = new StringBuilder(128);
		Node node = aNode;
		while(node!=null)
		{   switch(node.getNodeType())
			{   case Node.ATTRIBUTE_NODE:
					result.insert(0,aNode.getNodeName());
					result.insert(0,"/@");
					Attr a=(Attr)node;
					node=a.getOwnerElement();
					break;
					
				case Node.ELEMENT_NODE:
					int countInParent=1;
					String name = node.getNodeName();
					for (Node n=node.getPreviousSibling(); n!=null; n=n.getPreviousSibling())
					{	if (n instanceof Element)
						{	if (name.equals(n.getNodeName()))
							{	countInParent++;
							}
						}
					}
					int indexInParent = countInParent;
					if (countInParent==1)
					{	for (Node n=node.getNextSibling(); n!=null; n=n.getNextSibling())
						{	if (n instanceof Element)
							{	if (name.equals(n.getNodeName()))
								{	countInParent++;
									break;
								}
							}
						}
					}
					
					if (countInParent>1)
					{   result.insert(0,']');
						result.insert(0,indexInParent);
						result.insert(0,'[');
					}
					result.insert(0,name);
					result.insert(0,'/');
					node=node.getParentNode();
					break;
					
				default:
					node=node.getParentNode();
					break;
			}
		}
		return new String(result);
	}
	
	public static String getText(Node aNode)
	{   StringBuffer result = null;
		String result2 = null;
		for (Node n=aNode.getFirstChild(); n!=null; n=n.getNextSibling())
		{   if (n.getNodeType()==Node.TEXT_NODE || n.getNodeType()==Node.CDATA_SECTION_NODE)
			{	String s = n.getNodeValue();
				if (result2==null)
				{	result2 = s;
				}
				else if (result==null)
				{	result = new StringBuffer(result2.length()+s.length());
					result.append(result2);
					result.append(s);
				}
				else
				{	result.append(s);
				}
			}
		}
		if (result==null)
		{	if (result2==null)
			{	result2="";
			}
		}
		else
		{	result2 = new String(result);
		}
		return result2;
	}
	
	public static String getText(Node aNode, boolean aIgnoreWhitespace, boolean aTrim)
	{
		StringBuffer result = null;
		String result2 = null;
		for (Node n = aNode.getFirstChild(); n != null; n = n.getNextSibling())
		{
			if (n.getNodeType() == Node.TEXT_NODE || n.getNodeType() == Node.CDATA_SECTION_NODE)
			{
				String s = n.getNodeValue();
				if (!aIgnoreWhitespace | !XMLUtils.isWhitespace(s))
				{	if (result2 == null)
					{	result2 = s;
					}
					else if (result == null)
					{	result = new StringBuffer(result2.length() + s.length());
						result.append(result2);
						result.append(s);
					}
					else
					{	result.append(s);
					}
				}
			}
		}
		if (result == null)
		{	if (result2 == null)
			{	result2 = "";
			}
		}
		else
		{	result2 = new String(result);
		}
		if (aTrim)
		{	result2=result2.trim();
		}
		return result2;
	}
	
	public static boolean isWhitespace(String aText)
	{	boolean isWhitespace=true;
		for (int i=aText.length()-1; i>=0; i--)
		{	char c=aText.charAt(i);
			if (c!=' ' && c!='\n' && c!='\r' && c!='\t')
			{	isWhitespace=false;
				break;
			}
		}
		return isWhitespace;
	}
	
	public static boolean isWhitespace(char[] ch, int start, int length)
	{	boolean isWhitespace=true;
		for (int i=start+length-1; i>=start; i--)
		{	char c=ch[i];
			if (c!=' ' && c!='\n' && c!='\r' && c!='\t')
			{	isWhitespace=false;
				break;
			}
		}
		return isWhitespace;
	}
	
	
	public static void setText(Element aElement, String aText)
	{   for (Node n=aElement.getFirstChild(); n!=null; )
		{   if (n.getNodeType()==Node.TEXT_NODE || n.getNodeType()==Node.CDATA_SECTION_NODE)
			{   Node oldn = n;
				n=n.getNextSibling();
				aElement.removeChild(oldn);
			}
			else
			{   n=n.getNextSibling();
			}
		}
		Text t = aElement.getOwnerDocument().createTextNode(aText);
		aElement.appendChild(t);
	}
	
	/**
	 * finds the next sibling element
	 */
	public static Element getNextSiblingElement(Node aNode)
	{   Element result=null;
		for (Node n=aNode.getNextSibling(); n!=null; n=n.getNextSibling())
		{   if (n instanceof Element)
			{   result=(Element)n;
				break;
			}
		}
		return result;
	}
	/**
	 * finds the first child element
	 */
	public static Element getFirstChildElement(Node aNode)
	{   Element result=null;
		for (Node n=aNode.getFirstChild(); n!=null; n=n.getNextSibling())
		{   if (n instanceof Element)
			{   result=(Element)n;
				break;
			}
		}
		return result;
	}
	
	public static Element getChildElementNamed(Node aNode, String aName)
	{	Element e = getFirstChildElement(aNode);
		Element result=null;
		while (e!=null)
		{	if (pleaseDOMGiveMeTheNameOf(e).equals(aName))
			{	result = e;
				break;
			}
			e = getNextSiblingElement(e);
		}
		return result;
	}
	
	
	
	/** try down, then across, not ( up-across ) */
	public static Element depthFirstTraversalNextElement(Node aNode) {
		
		Element result=getFirstChildElement(aNode);
		if (result==null)
		{   result = getNextSiblingElement(aNode);
		}
		return result;
	}
	
	/** try down, then across, then ( up-across )
	 * @param aNode next element after this node
	 * @param aRoot we will not traverse higher than this node (if null ignored)
	 */
	public static Element inOrderTraversalNext(Node aNode, Node aRoot) {
		Element result=getFirstChildElement(aNode);
		if (result==null)
		{  result = getNextSiblingElement(aNode); 
		}
		if (result==null)
		{   Node parent=aNode.getParentNode();
			while(parent!=aRoot && parent!=null && result==null)
			{	result = getNextSiblingElement(parent);
				parent=parent.getParentNode();
			}
		}
		return result;
	}
	
	/** Returns the document for a node regardless of type
	 * - why is DOM so inconsistent?
	 */
	public static Document getDocumentFor(Node aNode)
	{	Document result;
		if (aNode instanceof Document)
		{	result = (Document)aNode;
		}
		else
		{	result = aNode.getOwnerDocument();
		}
		return result;
	}
	
    /**
     * Creates a named element with text if it doesn't already exist
	 * otherwise it just sets the text below it.
     * @param aParent node to add element below
     * @param aName name of element
     * @param aValue text value below element
     */
    public static void appendTextedElement(Node aParent,String aName, String aValue)
    {   Document d;
		if (aParent instanceof Document)
		{	d=(Document)aParent;
		}
		else
		{	d = aParent.getOwnerDocument();
		}
		Element e=null;
		for (Node n=aParent.getFirstChild(); n!=null; n=n.getNextSibling())
		{	if (n.getNodeType()==Node.ELEMENT_NODE && n.getNodeName().equals(aName))
			{	e=(Element)n;
				break;
			}
		}
		if (e==null)
		{	e = d.createElement(aName);
	        aParent.appendChild(e);
		}
		setText(e, aValue);
    }    
	
	/*Clones a fragment or a whole Document*/
	public static Node safeDeepClone(Node aNode)
	{	Document d = (aNode instanceof Document)?(Document)aNode:aNode.getOwnerDocument();
		synchronized(d)
		{	return aNode.cloneNode(true);
		}
	}
	
	public static String pleaseDOMGiveMeTheNameOf(Node aNode)
	{	String result = aNode.getLocalName();
		if (result==null)
		{	result = aNode.getNodeName();
		}
		return result;
	}
	
	/** Write an XML element with text to the writer
	 */
	public static void write(Writer osw, String aName, String aValue) throws IOException
	{
		osw.write('<');
		osw.write(aName);
		osw.write('>');
		osw.write(aValue);
		osw.write("</");
		osw.write(aName);
		osw.write('>');
	}
	
	public static void writeAttribute(Writer osw, String aName, String aValue) throws IOException
	{
		osw.write(aName);
		osw.write("='");
		osw.write(aValue);
		osw.write("' ");
	}
	
	/** Write an escaped XML element with text to the writer
	 */
	public static void writeEscaped(Writer osw, String aName, String aValue) throws IOException
	{	write(osw,aName,escape(aValue));
	}
	
	public static void writeOpenTag(Writer osw, String aName) throws IOException
	{	osw.write('<');
		osw.write(aName);
		osw.write('>');
	}
	public static void writeCloseTag(Writer osw, String aName) throws IOException
	{	osw.write("</");
		osw.write(aName);
		osw.write('>');
	}
		
	public static int documentOrderCompare(Element e1, Element e2)
	{	if (e1==e2)
		{	return 0;
		}
		Element n=e1.getOwnerDocument().getDocumentElement();
		while (n!=null)
		{	
			if (n==e1)
			{	return -1;
			}
			if (n==e2)
			{	return 1;
			}
			n=XMLUtils.inOrderTraversalNext(n,null);
		}
		return 0;
		
	}
	
	private static void innerToXML(Result aResult, Node aNode, boolean aIndent, boolean aOmitDeclaration, String aEncoding)
	throws IOException, TransformerConfigurationException, TransformerException
	{	Properties format = new Properties();
		format.setProperty( OutputKeys.METHOD,"xml");
		format.setProperty( OutputKeys.ENCODING,aEncoding);
		format.setProperty( OutputKeys.OMIT_XML_DECLARATION, aOmitDeclaration ? "yes" : "no");
		format.setProperty( OutputKeys.INDENT, aIndent ? "yes" : "no");
		TransformerFactory f = TransformerFactory.newInstance();
		Transformer t = f.newTransformer();
		t.setOutputProperties(format);
		Source input = new DOMSource(aNode);
		t.transform(input, aResult);
	}
	
	/** serialise the given node recursively to a string
	 */
	public static String toXML(Node aNode, boolean aIndent, boolean aOmitDeclaration) throws IOException, TransformerConfigurationException, TransformerException
	{	StringWriter sw = new StringWriter(1024);
		Result output = new StreamResult(sw);
		innerToXML(output,aNode,aIndent,aOmitDeclaration,"UTF-8");
		return new String(sw.getBuffer());
	}
	
	public static IReadableBinaryStreamRepresentation toXML(Node aNode, boolean aIndent, boolean aOmitDeclaration, String aEncoding) throws IOException, TransformerConfigurationException, TransformerException
	{	ByteArrayOutputStream baos = new ByteArrayOutputStream(1024);
		Result output = new StreamResult(baos);
		innerToXML(output,aNode,aIndent,aOmitDeclaration,aEncoding);
		return new ByteArrayRepresentation(baos,aEncoding);
	}
	
	/** parse the document in the given input stream
	 */
	public static Document parse(InputStream aStream)
	throws IOException, SAXException, ParserConfigurationException
	{  
		Thread t=Thread.currentThread();
		ClassLoader cl=t.getContextClassLoader();
		t.setContextClassLoader(XMLUtils.class.getClassLoader());
		try
		{	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();	
			factory.setNamespaceAware(true);
			DocumentBuilder builder = factory.newDocumentBuilder();
			InputSource is = new InputSource(aStream);
			return builder.parse(is);
		}
		finally
		{	t.setContextClassLoader(cl);
		}
	}
}
